"""test for change email settings"""
import requests_mock
from unittest import TestCase
from urllib.parse import urljoin

from edx_api.client import EdxApi
from edx_api.email_settings import EmailSettings


class TestEmailSettings(TestCase):

    @classmethod
    def setUpClass(cls):
        edx_base_url = "http://edx.example.com"
        cls.api_url = urljoin(edx_base_url, EmailSettings.api_url)
        cls.client = EdxApi({"access_token": "foobar"}, cls.api_url)
        cls.email_settings = cls.client.email_settings
        cls.json = {"success": "true"}
        cls.course_id = "course_id"

    @requests_mock.mock()
    def test_course_emails_subscription(self, mocked_request):
        mocked_request.post(self.api_url, json=self.json)
        response = self.email_settings.subscribe(self.course_id)
        self.assertDictEqual(
            mocked_request.last_request.json(),
            {"course_id": self.course_id, "receive_emails": "on"}
        )
        self.assertTrue(response)

    @requests_mock.mock()
    def test_course_emails_unsubscription(self, mocked_request):
        mocked_request.post(self.api_url, json=self.json)
        response = self.email_settings.unsubscribe(self.course_id)
        self.assertDictEqual(
            mocked_request.last_request.json(), {"course_id": self.course_id}
        )
        self.assertTrue(response)



"""
edX Enrollment REST API client class
"""
from edx_api.constants import ENROLLMENT_MODE_AUDIT, ENROLLMENT_MODE_VERIFIED

try:
    from urlparse import urlparse, parse_qs
except ImportError:
    from urllib.parse import urlparse, parse_qs
from urllib.parse import urljoin

from .models import Enrollment, Enrollments


# pylint: disable=too-few-public-methods
class CourseEnrollments(object):
    """
    edX student enrollments client
    """

    enrollment_url = '/api/enrollment/v1/enrollment'
    enrollment_list_url = '/api/enrollment/v1/enrollments'

    def __init__(self, requester, base_url):
        """
        Args:
            requester (Requester): an authenticated objects for requests to edX
            base_url (str): string representing the base URL of an edX LMS instance
        """
        self.requester = requester
        self.base_url = base_url

    def _get_enrollments_list_page(self, params=None):
        """
        Submit request to retrieve enrollments list.

        Args:
            params (dict): Query parameters to use in the request. Valid parameters are:
                * course_id: Filters the result to course enrollments for the course
                    corresponding to the given course ID. The value must be URL encoded.
                    Optional.
                * username: username: List of comma-separated usernames. Filters the result to the
                    course enrollments of the given users. Optional.
        """
        req_url = urljoin(self.base_url, self.enrollment_list_url)
        resp = self.requester.get(req_url, params=params)
        resp.raise_for_status()
        resp_json = resp.json()
        results = resp_json['results']
        next_url_str = resp_json.get('next')
        cursor = None
        qstr_cursor = None
        if next_url_str:
            next_url = urlparse(next_url_str)
            qstr = parse_qs(next_url.query)
            qstr_cursor = qstr.get('cursor')

        if qstr_cursor and isinstance(qstr_cursor, list):
            cursor = qstr_cursor[0]

        return results, cursor

    def get_enrollments(self, course_id=None, usernames=None):
        """
        List all course enrollments.

        Args:
            course_id (str, optional): If used enrollments will be filtered to the specified
                course id.
            usernames (list, optional): List of usernames to filter enrollments.

        Notes:
            - This method returns an iterator to avoid going through the entire pagination at once.
            - The :class:`Enrollments` instance returned for each generated item will not have any
                course details.

        Examples:
            Get all enrollments for a specific course id
            >>> api = EdxApi({'access_token': 'token'}, 'http://base_url')
            >>> enrollments = api.enrollments.get_enrollments(course_id='course_id')
            >>> for enrollment in enrollments:
                    do_something(enrollment)

            Get all enrollments for a set of usernames
            >>> api = EdxApi({'access_token': 'token'}, 'http://base_url')
            >>> enrollments = api.enrollments.get_enrollments(usernames=['user1', 'user2'])
            >>> for enrollment in enrollments:
                    do_something(enrollment)

        Returns:
            Generator with an instance of :class:`Enrollments` for each item.
        """
        params = {}
        if course_id is not None:
            params['course_id'] = course_id
        if usernames is not None and isinstance(usernames, list):
            params['username'] = ','.join(usernames)

        done = False
        while not done:
            enrollments, next_cursor = self._get_enrollments_list_page(params)
            for enrollment in enrollments:
                yield Enrollment(enrollment)

            if next_cursor:
                params['cursor'] = next_cursor
            else:
                done = True

    def get_student_enrollments(self):
        """
        Returns an Enrollments object with the user enrollments for the user
        whose access token was provided to the API client.

        Returns:
            Enrollments: object representing the student enrollments
        """
        resp = self.requester.get(
            urljoin(self.base_url, self.enrollment_url))
        resp.raise_for_status()
        return Enrollments(resp.json())

    def create_student_enrollment(
            self,
            course_id,
            mode=ENROLLMENT_MODE_AUDIT,
            username=None,
            enrollment_attributes=None
    ):
        """
        Creates an enrollment for the user in a given course

        Args:
            course_id (str): an edX course id.
            mode (str): Enrollment mode.
            username (str): Username.
            enrollment_attributes (dict): Enrollment attributes, which are added directly to
                                            the request body.
                                          If specified, the following keys are expected:
                                          - namespace,
                                          - name,
                                          - value.

        Returns:
            Enrollment: object representing the student enrollment in the provided course
        """
        enrollment_data = {
            "mode": mode,
            "course_details": {"course_id": course_id}
        }
        if username:
            enrollment_data['user'] = username
        if enrollment_attributes:
            enrollment_data['enrollment_attributes'] = enrollment_attributes
        # the request is done in behalf of the current logged in user
        resp = self.requester.post(
            urljoin(self.base_url, self.enrollment_url),
            json=enrollment_data
        )
        resp.raise_for_status()
        return Enrollment(resp.json())

    def create_audit_student_enrollment(self, course_id, username=None):
        """
        Creates an audit enrollment for the user in a given course

        Args:
            course_id (str): An edX course id.
            username (str): Username.

        Returns:
            Enrollment: object representing the student enrollment in the provided course
        """
        return self.create_student_enrollment(
            course_id,
            mode=ENROLLMENT_MODE_AUDIT,
            username=username
        )

    def create_verified_student_enrollment(self, course_id, username=None):
        """
        Creates a Verified enrollment for the user in a given course

        Args:
            course_id (str): An edX course id.
            username (str): Username.

        Returns:
            Enrollment: object representing the student enrollment in the provided course
        """
        # Note: The Verified enrollment can be created only by Staff or any user that has API keys level access.
        # So this function should only be called on behalf of a Staff user
        return self.create_student_enrollment(
            course_id,
            mode=ENROLLMENT_MODE_VERIFIED,
            username=username
        )

    def deactivate_enrollment(self, course_id):
        """
        Deactivates an enrollment in the given course for the user
        whose access token was provided to the API client (in other words, the
        user will be unenrolled)

        Args:
            course_id (str): An edX course id.

        Returns:
            Enrollment: object representing the deactivated student enrollment
        """
        resp = self.requester.post(
            urljoin(self.base_url, self.enrollment_url),
            json={
                "course_details": {"course_id": course_id},
                "is_active": False,
            }
        )
        resp.raise_for_status()
        return Enrollment(resp.json())



"""Course Detail API"""
from urllib.parse import urljoin

from .models import CourseDetail, CourseMode


# pylint: disable=too-few-public-methods
class CourseDetails(object):
    """
    API Client to interface with the course detail API.
    """

    def __init__(self, requester, base_url):
        self._requester = requester
        self._base_url = base_url

    def get_detail(self, course_id, username=None):
        """
        Fetches course details.

        Args:
            course_id (str): An edx course id.

        Returns:
            CourseDetail
        """
        # the request is done on behalf of the current logged in user
        # this only works if COURSE_ABOUT_VISIBILITIY_PERMISSION is not
        # set to staff, otherwise you need to pass in a username with
        # permissions.
        if not username:
            resp = self._requester.get(
                urljoin(
                    self._base_url,
                    "/api/courses/v1/courses/{course_key}".format(course_key=course_id),
                )
            )
        else:
            resp = self._requester.get(
                urljoin(
                    self._base_url,
                    "/api/courses/v1/courses/{course_key}/"
                    "?username={username}".format(
                        course_key=course_id, username=username
                    ),
                )
            )

        resp.raise_for_status()

        return CourseDetail(resp.json())


class CourseModes(object):
    """
    API Client to interface with the course modes API.
    """

    def __init__(self, requester, base_url):
        self._requester = requester
        self._base_url = base_url

    def get_mode(self, course_id):
        """
        Fetches course mode details.

        Args:
            course_id (str): An edx course id.

        Returns:
            List of CourseMode
        """
        resp = self._requester.get(
            urljoin(
                self._base_url,
                "/api/course_modes/v1/courses/{course_key}".format(
                    course_key=course_id
                ),
            )
        )

        resp.raise_for_status()
        course_mode_list = []
        for course_mode_json in resp.json():
            course_mode_list.append(CourseMode(course_mode_json))
        return course_mode_list



"""Tests for Info"""
import json
import os
from unittest import TestCase

from .models import Info


class InfoTests(TestCase):
    """Tests for Info"""

    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/user_info.json')) as file_obj:
            cls.user_info_json = json.loads(file_obj.read())

        cls.info = Info(cls.user_info_json)

    def test_str(self):
        """Test the __str__"""
        assert str(self.info) == "<User info for user staff>"

    def test_properties(self):
        """Test properties on Info model"""
        assert self.info.name == self.user_info_json['name']
        assert self.info.username == self.user_info_json['username']
        assert self.info.email == self.user_info_json['email']
        assert self.info.user_id == self.user_info_json['id']

    def test_missing(self):  # pylint: disable=no-self-use
        """Missing properties shouldn't cause a problem"""
        info = Info({})

        assert info.name is None
        assert info.username is None
        assert info.email is None
        assert info.user_id is None



"""
Business objects for the certificates API
"""
from dateutil import parser


class Certificates(object):
    """
    The certificates object
    This assumes that there can be only one certificate
    for a given course run and user.
    """
    def __init__(self, certificate_list):
        self.certificates = {}
        self.verified_certificates = {}
        for certificate in certificate_list:
            if not isinstance(certificate, Certificate):
                raise ValueError("Only Certificate objects are allowed")
            self.certificates[certificate.course_id] = certificate
            if certificate.is_verified:
                self.verified_certificates[certificate.course_id] = certificate

    def __str__(self):
        return "<Certificates>"

    @property
    def all_courses_certs(self):
        """Helper property to return all the course ids of the certificates"""
        return self.certificates.keys()

    @property
    def all_courses_verified_certs(self):
        """Helper property to return all the course ids of the verified certificates"""
        return self.verified_certificates.keys()

    @property
    def all_certs(self):
        """Helper property to return all the certificates"""
        return self.certificates.values()

    @property
    def all_verified_certs(self):
        """Helper property to return all the verified certificates"""
        return self.verified_certificates.values()

    def get_cert(self, course_id):
        """Returns the certificate for the given course id"""
        return self.certificates.get(course_id)

    def get_verified_cert(self, course_id):
        """Returns the verified certificate for the given course id"""
        return self.verified_certificates.get(course_id)

    def has_verified_cert(self, course_id):
        """Whether the course has a verified certificate"""
        return course_id in self.verified_certificates


class Certificate(object):
    """
    The certificate object
    """
    def __init__(self, json):
        self.json = json

    def __str__(self):
        return "<Certificate for user {username} for course {course_id}>".format(
            username=self.username,
            course_id=self.course_id
        )

    @property
    def is_verified(self):
        """Whether certificate_type is verified"""
        return self.certificate_type == 'verified'

    @property
    def username(self):
        """Returns the username property"""
        return self.json.get("username")

    @property
    def course_id(self):
        """Returns the course_id property"""
        return self.json.get("course_id")

    @property
    def certificate_type(self):
        """Returns the certificate_type property"""
        return self.json.get("certificate_type")

    @property
    def status(self):
        """Returns the status property"""
        return self.json.get("status")

    @property
    def download_url(self):
        """Returns the download_url property"""
        return self.json.get("download_url")

    @property
    def grade(self):
        """Returns the grade property"""
        return self.json.get("grade")

    @property
    def created(self):
        """Returns the created property"""
        return parser.parse(self.json.get('created'))

    @property
    def modified(self):
        """Returns the modified property"""
        return parser.parse(self.json.get('modified'))

    @property
    def is_passing(self):
        """Returns the is_passing property"""
        return self.json.get("is_passing")



"""Models Tests for the Certificates API"""

from datetime import datetime
import os.path
import json
from unittest import TestCase

from .models import (
    Certificate,
    Certificates,
)


class CertificatesTests(TestCase):
    """Tests for certificates object"""
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/certificates.json')) as file_obj:
            cls.certs_json = json.loads(file_obj.read())

        cls.certificates = Certificates([Certificate(cert_json) for cert_json in cls.certs_json])

    def test_str(self):
        """Test the __str__"""
        assert str(self.certificates) == "<Certificates>"

    def test_only_certificates(self):
        """Test that only certificates are allowed in the constructor"""
        with self.assertRaises(ValueError):
            Certificates(['foo', 'bar'])
        with self.assertRaises(ValueError):
            Certificates([{'foo': 'bar'}])

    def test_courses_cert(self):
        """Test for all_courses_certs"""
        assert sorted(list(self.certificates.all_courses_certs)) == sorted(
            ["course-v1:edX+DemoX+Demo_Course", "course-v1:MITx+8.MechCX+2014_T1"])

    def test_courses_verified_cert(self):
        """Test for all_courses_verified_certs"""
        assert list(self.certificates.all_courses_verified_certs) == [
            "course-v1:edX+DemoX+Demo_Course"
        ]

    def test_all_cert(self):
        """Test for all_certs"""
        all_certs = self.certificates.all_certs
        assert len(all_certs) == 2
        for cert in all_certs:
            assert isinstance(cert, Certificate)

    def test_all_verified_cert(self):
        """Test for all_verified_certs"""
        all_certs = list(self.certificates.all_verified_certs)
        assert len(all_certs) == 1
        assert isinstance(all_certs[0], Certificate)
        assert all_certs[0].course_id == "course-v1:edX+DemoX+Demo_Course"

    def test_get_certificate(self):
        """Test for get_cert"""
        assert self.certificates.get_cert('foo_id') is None
        assert isinstance(
            self.certificates.get_cert("course-v1:MITx+8.MechCX+2014_T1"),
            Certificate
        )

    def test_get_verified_cert(self):
        """Test for get_verified_cert"""
        assert self.certificates.get_verified_cert('foo_id') is None
        assert self.certificates.get_verified_cert("course-v1:MITx+8.MechCX+2014_T1") is None
        cert = self.certificates.get_verified_cert("course-v1:edX+DemoX+Demo_Course")
        assert isinstance(cert, Certificate)
        assert cert.is_verified is True

    def test_has_verified_cert(self):
        """Test for has_verified_cert"""
        assert self.certificates.has_verified_cert('foo_id') is False
        assert self.certificates.has_verified_cert(
            "course-v1:MITx+8.MechCX+2014_T1") is False
        assert self.certificates.has_verified_cert("course-v1:edX+DemoX+Demo_Course") is True


class CertificateTests(TestCase):
    """Tests for certificate object"""
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/certificates.json')) as file_obj:
            cls.certs_json = json.loads(file_obj.read())

        cls.certificate = Certificate(cls.certs_json[0])

    def test_str(self):
        """Test the __str__"""
        assert str(self.certificate) == ("<Certificate for user bob for "
                                         "course course-v1:edX+DemoX+Demo_Course>")

    def test_is_verified(self):
        """Test for is_verified"""
        assert self.certificate.is_verified is True

    def test_username(self):
        """Test for username"""
        assert self.certificate.username == "bob"

    def test_course_id(self):
        """Test for course_id"""
        assert self.certificate.course_id == "course-v1:edX+DemoX+Demo_Course"

    def test_certificate_type(self):
        """Test for certificate_type"""
        assert self.certificate.certificate_type == "verified"

    def test_status(self):
        """Test for status"""
        assert self.certificate.status == "downloadable"

    def test_download_url(self):
        """Test for download_url"""
        assert self.certificate.download_url == "http://www.example.com/demo.pdf"

    def test_grade(self):
        """Test for grade"""
        assert self.certificate.grade == "0.98"

    def test_is_passing(self):
        """Test for is_passing"""
        assert self.certificate.is_passing is True

    def test_created(self):
        """Test for created"""
        expected = datetime.strptime(
            "2015-07-31T00:00:00", "%Y-%m-%dT%H:%M:%S"
        )
        assert self.certificate.created == expected

    def test_modified(self):
        """Test for modified"""
        expected = datetime.strptime(
            "2015-08-31T00:00:00", "%Y-%m-%dT%H:%M:%S"
        )
        assert self.certificate.modified == expected



"""client tests"""
import pytest

from .client import EdxApi


def test_request_id_credential():
    """access_token required in credentials"""
    with pytest.raises(AttributeError) as exc:
        EdxApi({})

    assert 'access token' in str(exc.value)


def test_instantiation_happypath():
    """instantiatable with correct args"""
    token = 'asdf'
    api_key = 'api_key'
    client = EdxApi({'access_token': token, 'api_key': api_key})
    assert client.get_requester().headers['Authorization'] == 'Bearer {token}'.format(token=token)
    assert client.get_requester().headers['X-EdX-Api-Key'] == api_key



"""Models Tests for the Enrollment API"""

import os.path
import json
from unittest import TestCase

from dateutil import parser

from .models import (
    Enrollments,
    Enrollment,
    CourseDetails,
    CourseMode,
)


class EnrollmentsTests(TestCase):
    """Tests for enrollments object"""
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/user_enrollments.json')) as file_obj:
            cls.enrollments_json = json.loads(file_obj.read())

        cls.enrollments = Enrollments(cls.enrollments_json)

    def test_str(self):
        """Test the __str__"""
        assert str(self.enrollments) == "<Enrollments>"

    def test_enrollments(self):
        """Test the enrolled_courses method"""
        enrollments_list = list(self.enrollments.enrolled_courses)
        assert len(enrollments_list) == 2

        for elem in enrollments_list:
            assert isinstance(elem, Enrollment)

    def test_filter_enrolled_courses(self):
        """Tests helper method"""
        with self.assertRaises(ValueError):
            list(self.enrollments.get_enrolled_courses_by_ids('foo'))

        enrollments_list = list(
            self.enrollments.get_enrolled_courses_by_ids(('course-v1:edX+DemoX+Demo_Course',))
        )
        assert len(enrollments_list) == 1
        assert enrollments_list[0].course_id == 'course-v1:edX+DemoX+Demo_Course'

    def test_get_id_only(self):
        """Test for get_enrolled_course_ids method"""
        enrollments_id_list = list(
            self.enrollments.get_enrolled_course_ids()
        )
        assert len(enrollments_id_list) == 2

        enrollments_id_list = list(
            self.enrollments.get_enrolled_course_ids(('course-v1:edX+DemoX+Demo_Course',))
        )
        assert len(enrollments_id_list) == 1
        assert enrollments_id_list[0] == 'course-v1:edX+DemoX+Demo_Course'

        enrollments_id_list = list(
            self.enrollments.get_enrolled_course_ids(('course-v1:edX+DemoX+Demo_Course', 'foo'))
        )
        assert len(enrollments_id_list) == 1
        assert enrollments_id_list[0] == 'course-v1:edX+DemoX+Demo_Course'

    def test_is_enrolled_in(self):
        """Test for is_enrolled_in"""
        assert self.enrollments.is_enrolled_in('course-v1:foo+bar+baz_course') is False
        assert self.enrollments.is_enrolled_in('course-v1:edX+DemoX+Demo_Course') is True

    def test_get_enrollment_for_course(self):
        """Test for get_enrollment_for_course"""
        assert self.enrollments.get_enrollment_for_course('foo') is None
        enr = self.enrollments.get_enrollment_for_course('course-v1:edX+DemoX+Demo_Course')
        assert isinstance(enr, Enrollment)


class EnrollmentTests(TestCase):
    """Tests for enrollment object"""
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/user_enrollments.json')) as file_obj:
            cls.enrollments_json = json.loads(file_obj.read())

        cls.enrollment_json = cls.enrollments_json[0]
        cls.enrollment = Enrollment(cls.enrollment_json)

    def test_str(self):
        """Test the __str__"""
        assert str(self.enrollment) == ("<Enrollment for user staff in "
                                        "course course-v1:edX+DemoX+Demo_Course>")

    def test_created(self):
        """Test for created property"""
        assert self.enrollment.created == parser.parse("2015-12-22T02:47:17Z")

    def test_mode(self):
        """Test for mode property"""
        assert self.enrollment.mode == "honor"

    def test_is_active(self):
        """Test for is_active property"""
        assert self.enrollment.is_active is True

    def test_course_details(self):
        """Test for course_details property"""
        assert isinstance(self.enrollment.course_details, CourseDetails)

    def test_user(self):
        """Test for user property"""
        assert self.enrollment.user == "staff"

    def test_course_id(self):
        """Test for course id property"""
        assert self.enrollment.course_id == "course-v1:edX+DemoX+Demo_Course"
        assert self.enrollment.course_id == self.enrollment.course_details.course_id

    def test_is_verified(self):
        """Test for is_active property"""
        assert self.enrollment.is_verified is False


class CourseDetailsTests(TestCase):
    """Tests for course details object"""
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/user_enrollments.json')) as file_obj:
            cls.enrollments_json = json.loads(file_obj.read())

        cls.course_details_json = cls.enrollments_json[0]['course_details']
        cls.course_details = CourseDetails(cls.course_details_json)

    def test_str(self):
        """Test the __str__"""
        assert str(self.course_details) == ("<Enrollment details for course "
                                            "course-v1:edX+DemoX+Demo_Course>")

    def test_course_id(self):
        """Test for course_id property"""
        assert self.course_details.course_id == "course-v1:edX+DemoX+Demo_Course"

    def test_course_start(self):
        """Test for course_start property"""
        assert self.course_details.course_start == parser.parse("2013-02-05T05:00:00Z")

    def test_course_end(self):
        """Test for course_end property"""
        assert self.course_details.course_end == parser.parse("2015-05-10T23:59:00Z")

    def test_enrollment_start(self):
        """Test for enrollment_start property"""
        assert self.course_details.enrollment_start == parser.parse("2013-02-05T05:00:01Z")

    def test_enrollment_end(self):
        """Test for enrollment_end property"""
        assert self.course_details.enrollment_end == parser.parse("2015-05-10T23:58:00Z")

    def test_invite_only(self):
        """Test for invite_only property"""
        assert self.course_details.invite_only is False

    def test_course_modes(self):
        """Test for course_modes property"""
        course_modes_list = list(self.course_details.course_modes)
        assert len(course_modes_list) == 1
        for elem in course_modes_list:
            assert isinstance(elem, CourseMode)


class CourseModeTests(TestCase):
    """Tests for course mode object"""
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/user_enrollments.json')) as file_obj:
            cls.enrollments_json = json.loads(file_obj.read())

        cls.course_mode_json = cls.enrollments_json[0]['course_details']['course_modes'][0]
        cls.course_mode = CourseMode(cls.course_mode_json)

    def test_currency(self):
        """Test for currency property"""
        assert self.course_mode.currency == "usd"

    def test_description(self):
        """Test for description property"""
        assert self.course_mode.description is None

    def test_expiration_datetime(self):
        """Test for expiration_datetime property"""
        assert self.course_mode.expiration_datetime is None

    def test_min_price(self):
        """Test for min_price property"""
        assert self.course_mode.min_price == 0

    def test_name(self):
        """Test for name property"""
        assert self.course_mode.name == "Audit"

    def test_slug(self):
        """Test for slug property"""
        assert self.course_mode.slug == "audit"

    def test_suggested_prices(self):
        """Test for  property"""
        assert self.course_mode.suggested_prices == ""



"""
edX Grades REST API client class
"""
from requests.exceptions import HTTPError
from urllib.parse import urljoin

from edx_api.enrollments import CourseEnrollments
from .models import CurrentGrade, CurrentGradesByUser, CurrentGradesByCourse


class UserCurrentGrades(object):
    """
    edX student certificates client
    """

    def __init__(self, requester, base_url):
        """
        Args:
            requester (Requester): an authenticated objects for requests to edX
            base_url (str): string representing the base URL of an edX LMS instance
        """
        self.requester = requester
        self.base_url = base_url

    def get_student_current_grade(self, username, course_id):
        """
        Returns an CurrentGrade object for the user in a course

        Args:
            username (str): an edx user's username
            course_id (str): an edX course id.

        Returns:
            CurrentGrade: object representing the student current grade for a course
        """
        # the request is done in behalf of the current logged in user
        resp = self.requester.get(
            urljoin(
                self.base_url,
                '/api/grades/v1/courses/{course_key}/?username={username}'.format(
                    username=username,
                    course_key=course_id
                )
            )
        )

        resp.raise_for_status()

        return CurrentGrade(resp.json()[0])

    def get_student_current_grades(self, username, course_ids=None):
        """
        Returns a CurrentGradesByUser object with the user current grades.

        Args:
            username (str): an edx user's username
            course_ids (list): a list of edX course ids.

        Returns:
            CurrentGradesByUser: object representing the student current grades
        """
        # if no course ids are provided, let's get the user enrollments
        if course_ids is None:
            enrollments_client = CourseEnrollments(self.requester, self.base_url)
            enrollments = enrollments_client.get_student_enrollments()
            course_ids = list(enrollments.get_enrolled_course_ids())

        all_current_grades = []
        for course_id in course_ids:
            try:
                all_current_grades.append(self.get_student_current_grade(username, course_id))
            except HTTPError as error:
                if error.response.status_code >= 500:
                    raise

        return CurrentGradesByUser(all_current_grades)

    def get_course_current_grades(self, course_id):
        """
        Returns a CurrentGradesByCourse object for all users in the specified course.

        Args:
            course_id (str): an edX course ids.

        Returns:
            CurrentGradesByCourse: object representing the student current grades

        Authorization:
            The authenticated user must have staff permissions to see grades for all users
            in a course.
        """
        resp = self.requester.get(
            urljoin(
                self.base_url,
                '/api/grades/v1/courses/{course_key}/'.format(course_key=course_id)
            )
        )
        resp.raise_for_status()
        resp_json = resp.json()
        if 'results' in resp_json:
            grade_entries = [CurrentGrade(entry) for entry in resp_json["results"]]
            while resp_json['next'] is not None:
                resp = self.requester.get(resp_json['next'])
                resp.raise_for_status()
                resp_json = resp.json()
                grade_entries.extend((CurrentGrade(entry) for entry in resp_json["results"]))
        else:
            grade_entries = [CurrentGrade(entry) for entry in resp_json]

        return CurrentGradesByCourse(grade_entries)



"""Models Tests for the Grades API"""

import json
import os.path
from copy import deepcopy
from unittest import TestCase

from .models import (
    CurrentGrade,
    CurrentGradesByCourse,
    CurrentGradesByUser,
)


class CurrentGradesByCourseTests(TestCase):
    """
    Tests for CurrentGradesByCourse
    """
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/course_grades_hawthorn.json')) as file_obj:
            cls.grades_json = json.loads(file_obj.read())

        cls.current_grades = CurrentGradesByCourse(
            [CurrentGrade(json_obj) for json_obj in cls.grades_json]
        )

    def test_expected_iterable(self):
        """CurrentGradesByCourse expects an iterable as input"""
        with self.assertRaises(TypeError):
            CurrentGradesByCourse(123)

    def test_cgbycourse_objects(self):
        """CurrentGradesByCourse can contain only CurrentGrade objects"""
        with self.assertRaises(ValueError):
            CurrentGradesByCourse([{'foo': 'bar'}])

    def test_only_same_course_grades(self):
        """
        CurrentGradesByCourse should only contain grades from a single course_id.
        """
        grades_json = deepcopy(self.grades_json)
        grades_json[0]['course_id'] = "course-v1:A:Nonreal:Course"
        with self.assertRaises(ValueError):
            CurrentGradesByCourse(
                [CurrentGrade(json_obj) for json_obj in grades_json],
            )

    def test_different_usernames(self):
        """Grades can have different users."""
        grades_json = deepcopy(self.grades_json)
        for grade in grades_json:
            grade['course_id'] = grades_json[0]['course_id']
        grades_json[0]['username'] = 'different_username'
        # No exception raised
        CurrentGradesByCourse(
            [CurrentGrade(json_obj) for json_obj in grades_json],
        )

    def test_all_current_grades(self):
        """Test for all_current_grades property"""
        all_grades = self.current_grades.all_current_grades
        assert len(all_grades) == 2
        for grade in all_grades:
            assert isinstance(grade, CurrentGrade)


class CurrentGradesByUserTests(TestCase):
    """
    Tests for CurrentGradesByUser
    """
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/current_grades.json')) as file_obj:
            cls.grades_json = json.loads(file_obj.read())

        cls.current_grades = CurrentGradesByUser(
            [CurrentGrade(json_obj) for json_obj in cls.grades_json]
        )

    def test_expected_iterable(self):
        """CurrentGradesByUser expects an iterable as input"""
        with self.assertRaises(TypeError):
            CurrentGradesByUser(123)

    def test_cgbyuser_objects(self):
        """CurrentGradesByUser can contain only CurrentGrade objects"""
        with self.assertRaises(ValueError):
            CurrentGradesByUser([{'foo': 'bar'}])

    def test_all_course_ids(self):
        """Test for all_course_ids property"""
        assert set(self.current_grades.all_course_ids) == {
            "course-v1:edX+DemoX+Demo_Course", "course-v1:MITx+8.MechCX+2014_T1"
        }

    def test_username_is_different(self):
        """CurrentGradesByUser can contain grades for the user with different username"""
        grades_json = deepcopy(self.grades_json)
        grades_json[0]['username'] = 'other_random_string'
        CurrentGradesByUser([CurrentGrade(json_obj) for json_obj in grades_json])

    def test_get_current_grade(self):
        """Test for get_current_grade method"""
        course_grade = self.current_grades.get_current_grade("course-v1:edX+DemoX+Demo_Course")
        assert isinstance(course_grade, CurrentGrade)
        assert course_grade.course_id == "course-v1:edX+DemoX+Demo_Course"

    def test_all_current_grades(self):
        """Test for all_current_grades property"""
        all_grades = self.current_grades.all_current_grades
        assert len(all_grades) == 2
        for grade in all_grades:
            assert isinstance(grade, CurrentGrade)


class CurrentGradeTests(TestCase):
    """Tests for current grade object"""
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/current_grades.json')) as file_obj:
            cls.grades_json = json.loads(file_obj.read())

        cls.current_grade = CurrentGrade(cls.grades_json[0])

    def test_str(self):
        """Test the __str__"""
        assert str(self.current_grade) == ("<Current Grade for user bob in "
                                           "course course-v1:edX+DemoX+Demo_Course>")

    def test_course_id(self):
        """Test for course_id property"""
        assert self.current_grade.course_id == "course-v1:edX+DemoX+Demo_Course"

    def test_email(self):
        """Test for email property"""
        assert self.current_grade.email == "bob@example.com"

    def test_username(self):
        """Test for user property"""
        assert self.current_grade.username == "bob"

    def test_passed(self):
        """Test for passed property"""
        assert self.current_grade.passed is True

    def test_percent(self):
        """Test for percent property"""
        assert self.current_grade.percent == 0.97

    def test_letter_grade(self):
        """Test for letter_grade property"""
        assert self.current_grade.letter_grade == "Pass"



"""
Integration tests.

These run against actual edX APIs.

NOTE: To run this test you need a running instance of edX properly configured.
This means that you need:
- the edx demo course set to course mode 'audit' (the default)
- the user staff being "staff" of such course
- the course being open to enrollment
- the course having the ccx enabled in the advanced settings
- a valid access token for the user "staff"
- another course is available with enrollment open and "staff" NOT enrolled
- the user staff is enrolled in, at least, two courses
- make sure LMS's settings.EDX_API_TOKEN value is the same as the token used here
- make sure user staff is admin in the demo course
- you run the following code in a python shell inside your devstack
  instance to create a certificate:
```
from certificates.models import CertificateStatuses
from django.contrib.auth.models import User
from certificates.tests.factories import GeneratedCertificateFactory
from opaque_keys.edx.keys import CourseKey
course_key = CourseKey.from_string('course-v1:edX+DemoX+Demo_Course')
staff = User.objects.get(username='staff')
GeneratedCertificateFactory.create(
    user=staff,
    course_id=course_key,
    status=CertificateStatuses.downloadable,
    mode='verified',
    download_url='www.google.com', grade="0.88",
)
```
"""
import os

from mock import patch
import pytest
from requests.exceptions import HTTPError
from requests import Response, Timeout
from urllib.parse import urljoin

from edx_api.constants import ENROLLMENT_MODE_AUDIT, ENROLLMENT_MODE_VERIFIED
from edx_api.client import EdxApi


ACCESS_TOKEN = os.getenv('ACCESS_TOKEN')
API_KEY = os.getenv('API_KEY')
BASE_URL = os.getenv('BASE_URL', 'http://localhost:18000/')
ENROLLMENT_CREATION_COURSE_ID = os.getenv('ENROLLMENT_CREATION_COURSE_ID')

# pylint: disable=invalid-name
require_integration_settings = pytest.mark.skipif(
    not ACCESS_TOKEN or not BASE_URL,
    reason="You must specify the credential env vars"
)

require_integration_settings_course_id = pytest.mark.skipif(
    not ACCESS_TOKEN or not BASE_URL or not ENROLLMENT_CREATION_COURSE_ID,
    reason="You must specify the credential env vars and the alternate course id for enrollments"
)


class FakeErroredResponse(object):
    """Fake requests response"""
    def __init__(self, status_code):
        """
        Build a fake error response
        """
        self.error = HTTPError(response=Response())
        self.error.response = Response()
        self.error.response.status_code = status_code

    def raise_for_status(self):
        """Raise an HTTPError"""
        raise self.error


class FakeTimeoutResponse(object):
    """Fake requests response"""
    def __init__(self, status_code):
        """
        Build a fake error response
        """
        self.error = Timeout(response=Response())
        self.error.response = Response()
        self.error.response.status_code = status_code

    def raise_for_status(self):
        """Raise an HTTPError"""
        raise self.error

    @staticmethod
    def json():
        """dummy error json"""
        return {
            "message": "time out"
        }


def mocked_timeout(url, *args, **kwargs):  # pylint: disable=unused-argument
    """
    Return a time out error.
    """
    return FakeTimeoutResponse(status_code=408)


@require_integration_settings
def test_course_structure():
    """
    Pull down the course structure and validate it has the correct entries.
    This test assumes that the used user can access the course.
    """
    api = EdxApi({'access_token': ACCESS_TOKEN, 'api_key': API_KEY}, base_url=BASE_URL)
    structure = api.course_structure.course_blocks('course-v1:edX+DemoX+Demo_Course', 'staff')

    titles = [
        block.title for block in structure.root.children
        if block.visible
    ]

    assert titles == [
        'Introduction',
        'Example Week 1: Getting Started',
        'Example Week 2: Get Interactive',
        'Example Week 3: Be Social',
        'About Exams and Certificates',
        'holding section',
    ]


@require_integration_settings
def test_enrollments():
    """
    Enrolls the user in a course and then pulls down the enrollments for the user.
    This assumes that the course in the edX instance is available for enrollment.
    """
    api = EdxApi({'access_token': ACCESS_TOKEN, 'api_key': API_KEY}, base_url=BASE_URL)

    # the enrollment will be done manually until
    # a client to enroll the student is implemented
    requester = api.get_requester()
    requester.post(
        urljoin(BASE_URL, '/api/enrollment/v1/enrollment'),
        json={"course_details": {"course_id": 'course-v1:edX+DemoX+Demo_Course'}}
    )

    enrollments = api.enrollments.get_student_enrollments()

    enrolled_courses = [
        enrolled_course.course_details.course_id
        for enrolled_course in enrollments.enrolled_courses
    ]

    assert 'course-v1:edX+DemoX+Demo_Course' in enrolled_courses


@require_integration_settings_course_id
def test_create_verified_enrollment():
    """
    Integration test to enroll the user in a course with `verified` mode
    """
    api = EdxApi({'access_token': ACCESS_TOKEN, 'api_key': API_KEY}, base_url=BASE_URL)
    enrollment = api.enrollments.create_student_enrollment(
        course_id=ENROLLMENT_CREATION_COURSE_ID,
        mode=ENROLLMENT_MODE_VERIFIED,
        username="staff"
    )
    assert enrollment.course_id == ENROLLMENT_CREATION_COURSE_ID
    assert enrollment.mode == ENROLLMENT_MODE_VERIFIED


@require_integration_settings_course_id
def test_create_audit_enrollment():
    """
    Integration test to enroll the user in a course with `audit` mode
    """
    api = EdxApi({'access_token': ACCESS_TOKEN, 'api_key': API_KEY}, base_url=BASE_URL)
    enrollment = api.enrollments.create_student_enrollment(
        course_id=ENROLLMENT_CREATION_COURSE_ID,
        mode=ENROLLMENT_MODE_AUDIT,
        username="staff"
    )
    assert enrollment.course_id == ENROLLMENT_CREATION_COURSE_ID
    assert enrollment.mode == ENROLLMENT_MODE_AUDIT


@require_integration_settings_course_id
def test_deactivate_enrollment():
    """
    Integration test to enroll then deactivate a user in a course
    """
    api = EdxApi({'access_token': ACCESS_TOKEN, 'api_key': API_KEY}, base_url=BASE_URL)
    api.enrollments.create_student_enrollment(
        course_id=ENROLLMENT_CREATION_COURSE_ID,
        mode=ENROLLMENT_MODE_AUDIT
    )
    deactivated_enrollment = api.enrollments.deactivate_enrollment(
        course_id=ENROLLMENT_CREATION_COURSE_ID
    )
    assert deactivated_enrollment.course_id == ENROLLMENT_CREATION_COURSE_ID
    assert deactivated_enrollment.is_active is False


@require_integration_settings_course_id
def test_create_enrollment_timeout():
    """
     Asserts request timeout error on enrollment api
    """
    api = EdxApi({'access_token': ACCESS_TOKEN, 'api_key': API_KEY}, base_url=BASE_URL)
    enrollments = api.enrollments
    with patch.object(enrollments.requester, 'post', autospec=True) as post:
        post.side_effect = mocked_timeout
        with pytest.raises(Timeout):
            enrollments.create_student_enrollment(ENROLLMENT_CREATION_COURSE_ID)


@require_integration_settings
def test_course_details():
    """
    Pull down the course details.
    This test assumes that the used user is not anonymous.
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    details = api.course_detail.get_detail('course-v1:edX+DemoX+Demo_Course')

    assert details.course_id == "course-v1:edX+DemoX+Demo_Course"
    assert details.name == "edX Demonstration Course"


@require_integration_settings
def test_course_details_timeout():
    """
    assert timeout exception on get course_detail
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    course_detail = api.course_detail

    with patch.object(course_detail._requester, 'get', autospec=True) as get:   # pylint: disable=protected-access
        get.side_effect = mocked_timeout
        with pytest.raises(Timeout):
            course_detail.get_detail('course-v1:edX+DemoX+Demo_Course')


@require_integration_settings
def test_create_ccx():
    """
    Creates a CCX for the demo course. This course *MUST* have ccx enabled in
    the advanced settings.
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    ccx_id = api.ccx.create(
        'course-v1:edX+DemoX+Demo_Course',
        'verified@example.com',
        100,
        'My CCX from test_create_ccx integration test via edx-api-client.'
    )

    assert ccx_id is not None
    assert '@' in ccx_id  # follows ccx format


@require_integration_settings
def test_get_timeout_error_ccx():
    """
    Asserts request timeout error on ccx
    """

    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    ccx = api.ccx

    with patch.object(ccx.requester, 'post', autospec=True) as post:
        post.side_effect = mocked_timeout
        with pytest.raises(Timeout):
            ccx.create(
                'course-v1:edX+DemoX+Demo_Course',
                'verified@example.com',
                100,
                'My CCX from test_create_ccx integration test via edx-api-client.'
            )


@require_integration_settings
def test_get_certificate():
    """
    Gets the certificate for the demo course.
    See this module docstring for the code to run to create one
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    certificate = api.certificates.get_student_certificate(
        'staff', 'course-v1:edX+DemoX+Demo_Course')
    assert certificate.username == 'staff'
    assert certificate.is_verified is True

    certificates = api.certificates.get_student_certificates('staff')
    assert len(certificates.all_courses_certs) >= 1
    assert 'course-v1:edX+DemoX+Demo_Course' in certificates.all_courses_certs
    assert 'course-v1:edX+DemoX+Demo_Course' in certificates.all_courses_verified_certs


@require_integration_settings
def test_get_certificate_500_error():
    """
    Asserts that a 500 error returned from EDX will be propagated
    """

    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    username = 'staff'
    course_key = 'course-v1:edX+DemoX+Demo_Course'

    certificates = api.certificates
    old_requester_get = certificates.requester.get

    def mocked_get(url, *args, **kwargs):
        """
        Return an error for specific URLs
        """
        if '/api/certificates/v0/certificates/' in url:
            return FakeErroredResponse(status_code=500)
        return old_requester_get(url, *args, **kwargs)

    with patch.object(certificates.requester, 'get', autospec=True) as get:
        get.side_effect = mocked_get
        with pytest.raises(HTTPError):
            certificates.get_student_certificate(
                username, course_key
            )

        with pytest.raises(HTTPError):
            certificates.get_student_certificates(username)


@require_integration_settings
def test_get_certificate_timeout_error():
    """
    Asserts request timeout error on certificate apis
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    username = 'staff'
    course_key = 'course-v1:edX+DemoX+Demo_Course'
    certificates = api.certificates

    with patch.object(certificates.requester, 'get', autospec=True) as get:
        get.side_effect = mocked_timeout
        with pytest.raises(Timeout):
            certificates.get_student_certificate(
                username, course_key
            )

        with pytest.raises(Timeout):
            certificates.get_student_certificates(username)


@require_integration_settings
def test_get_certificate_404_error():
    """
    Asserts that a 404 returned from EDX will be silenced for get_student_certificates
    """

    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    username = 'staff'
    course_key = 'course-v1:edX+DemoX+Demo_Course'

    certificates = api.certificates
    old_requester_get = certificates.requester.get

    def mocked_get(url, *args, **kwargs):
        """
        Return an error for specific URLs
        """
        if '/api/certificates/v0/certificates/' in url:
            return FakeErroredResponse(status_code=404)
        return old_requester_get(url, *args, **kwargs)

    with patch.object(certificates.requester, 'get', autospec=True) as get:
        get.side_effect = mocked_get
        with pytest.raises(HTTPError):
            certificates.get_student_certificate(
                username, course_key
            )

        # Note no error here, just empty list
        certs = certificates.get_student_certificates(username)
        assert not certs.all_courses_certs


@require_integration_settings
def test_get_current_grade():
    """
    Gets the user current grade.
    If an user is enrolled in a course she has a current grade (probably with percent == 0.0)
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    course_grade = api.current_grades.get_student_current_grade(
        'staff', 'course-v1:edX+DemoX+Demo_Course')
    assert course_grade.username == 'staff'

    student_grades = api.current_grades.get_student_current_grades('staff')
    assert len(student_grades.all_course_ids) >= 1
    assert 'course-v1:edX+DemoX+Demo_Course' in student_grades.all_course_ids

    course_grades = api.current_grades.get_course_current_grades('course-v1:edX+DemoX+Demo_Course')
    assert len(course_grades.all_usernames) >= 1
    assert 'honor' in course_grades.all_usernames


@require_integration_settings
def test_get_current_grade_timeout():
    """
    assert timeout exception on get current grade.
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    current_grades = api.current_grades
    with patch.object(current_grades.requester, 'get', autospec=True) as get:
        get.side_effect = mocked_timeout
        with pytest.raises(Timeout):
            current_grades.get_student_current_grade('staff', 'course-v1:edX+DemoX+Demo_Course')


@require_integration_settings
def test_user_info():
    """
    Gets information about the logged in user
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    info = api.user_info.get_user_info()
    assert info.username == 'staff'
    assert info.name == ''
    assert info.email == 'staff@example.com'
    assert isinstance(info.user_id, int)


@require_integration_settings
def test_user_info_timeout():
    """
    assert timeout exception on user_info
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    user_info = api.user_info
    with patch.object(user_info.requester, 'get', autospec=True) as get:
        get.side_effect = mocked_timeout
        with pytest.raises(Timeout):
            user_info.get_user_info()


@require_integration_settings
def test_enrollments_list():
    """
    Enrolls the user in a course and then pulls down the enrollments for the user.
    This assumes that the course in the edX instance is available for enrollment.
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    enrollments = api.enrollments.get_enrollments()

    cnt = 0
    for enrollment in enrollments:
        assert enrollment.course_id
        assert enrollment.created
        assert enrollment.mode
        assert enrollment.is_active
        assert enrollment.user
        if enrollment.mode != ENROLLMENT_MODE_VERIFIED:
            assert not enrollment.is_verified
        else:
            assert enrollment.is_verified
        cnt += 1

    assert cnt >= 2


@require_integration_settings
def test_user_name_update():
    """
    Asserts that update user's name api updates the full name of the user correctly.
    """
    api = EdxApi({'access_token': ACCESS_TOKEN}, base_url=BASE_URL)
    user_name = 'Test Name'
    updated_user = api.user_info.update_user_name('staff', user_name)
    assert updated_user.username == 'staff'
    assert updated_user.name == user_name



"""Model Tests"""
import os.path
import json
from unittest import TestCase

from .models import Structure, Block


class StructureTests(TestCase):
    """Tests for structure object"""
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/course_structure.json')) as file_obj:
            cls.structure = json.loads(file_obj.read())

        cls.subject = Structure(cls.structure)

    def test_root_obj(self):
        """gets root object"""
        assert self.subject.root.block_id == self.structure['root']

    def test_blocks(self):
        """blocks return iterable of blocks"""
        blocks = self.subject.blocks
        assert isinstance(next(blocks), Block)

    def test_str(self):
        """str test"""
        assert str(self.subject) == (
            "Structure for block-v1:edX+DemoX+Demo_Course+type"
            "@course+block@course"
        )


BLOCK_ID = (
    "block-v1:edX+DemoX+Demo_Course+type@chapter+block"
    "@1414ffd5143b4b508f739b563ab468b7"
)


class BlockTests(TestCase):
    """Tests for block object"""
    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/course_structure.json')) as file_obj:
            cls.structure = json.loads(file_obj.read())

    def test_children_length(self):
        """Children returns the correct number of blocks"""
        block = Block(BLOCK_ID, self.structure)
        assert len(block.children) == 1

    def test_visiblity(self):
        """Visible returns correct result"""
        block = Block(BLOCK_ID, self.structure)
        assert block.visible

    def test_title(self):
        """Title returns correct result"""
        block = Block(BLOCK_ID, self.structure)
        assert block.title == "About Exams and Certificates"

    def test_str(self):
        """str test"""
        block = Block(BLOCK_ID, self.structure)
        assert str(block) == BLOCK_ID



"""
API Client for interacting w/ CCXs
"""
import logging
from urllib import parse


log = logging.getLogger(__name__)


class CCX(object):
    """
    API Client for interacting w/ CCXs
    """
    def __init__(self, requester, base_url):
        self.requester = requester
        self.base_url = base_url

    # pylint: disable=too-many-arguments
    def create(self, master_course_id, coach_email, max_students_allowed, title, modules=None):
        """
        Creates a CCX

        Args:
            master_course_id (str): edx course id of the master course
            coach_email (str): email of the user to make a coach. This user must exist on edx.
            max_students_allowed (int): Maximum number of students to allow in this ccx.
            title (str): Title of the CCX to be created
            modules (optional list): A list of locator_ids (str) for the modules to enable.

        Returns:
           ccx_id (str): The ID of the ccx.
        """
        payload = {
            'master_course_id': master_course_id,
            'coach_email': coach_email,
            'max_students_allowed': max_students_allowed,
            'display_name': title,
        }

        if modules is not None:
            payload['course_modules'] = modules

        resp = self.requester.post(
            parse.urljoin(self.base_url, '/api/ccx/v0/ccx/'),
            json=payload
        )

        try:
            resp.raise_for_status()
        except Exception:
            log.error(resp.json())
            raise

        return resp.json()['ccx_course_id']



"""
Tests for the content of the __init__ module
"""
import json
import os
from unittest import TestCase

try:
    from mock import patch
except ImportError:
    from unittest.mock import patch

import requests_mock
from urllib.parse import urljoin

from edx_api.client import EdxApi
from edx_api.constants import ENROLLMENT_MODE_AUDIT, ENROLLMENT_MODE_VERIFIED
from edx_api.enrollments import CourseEnrollments


class EnrollmentsTest(TestCase):
    """
    Tests for the enrollments base function in the __init__ module
    """

    @classmethod
    def setUpClass(cls):
        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/user_enrollments.json')) as file_obj:
            cls.enrollments_json = json.loads(file_obj.read())

        with open(os.path.join(os.path.dirname(__file__),
                               'fixtures/enrollments_list.json')) as file_obj:
            cls.enrollments_list_json = json.loads(file_obj.read())

        cls.enrollment_responses = [
            {'json': cls.enrollments_json[0], 'status_code': 200},
            {'json': cls.enrollments_json[1], 'status_code': 200},
        ]

        base_edx_url = 'http://edx.example.com'
        cls.enrollment_url = urljoin(base_edx_url, CourseEnrollments.enrollment_url)
        cls.client = EdxApi({'access_token': 'foobar'}, cls.enrollment_url)
        cls.enrollment_client = cls.client.enrollments

    @requests_mock.mock()
    def test_create_enrollment_value(self, mock_req):
        """
        Tests the post request to create an enrollment.
        This just tests that the client expects a JSON object representing the enrollment.
        """
        mock_req.register_uri('POST', self.enrollment_url, self.enrollment_responses)
        course_id = 'dummy_course_id'
        returned_enrollments = [
            self.enrollment_client.create_student_enrollment(course_id),
            self.enrollment_client.create_audit_student_enrollment(course_id),
        ]
        assert [enrollment.json for enrollment in returned_enrollments] == self.enrollments_json

    @requests_mock.mock()
    def test_create_enrollment_body(self, request_mock):
        """
        Tests the post body crafted to create an enrollment.
        """
        request_mock.post(self.enrollment_url, json=self.enrollments_json[0])
        course_id = 'course_id'
        user = 'user'
        enrollment_attributes = {
            'namespace': 'credit',
            'name': 'provider_id',
            'value': 'institution_name',
        }
        self.enrollment_client.create_student_enrollment(
            course_id=course_id,
            mode=ENROLLMENT_MODE_VERIFIED
        )
        self.assertDictEqual(
            request_mock.last_request.json(),
            {
                'course_details': {
                    'course_id': course_id
                },
                'mode': ENROLLMENT_MODE_VERIFIED
            }
        )
        self.enrollment_client.create_audit_student_enrollment(course_id=course_id)
        self.assertDictEqual(
            request_mock.last_request.json(),
            {
                'course_details': {
                    'course_id': course_id
                },
                'mode': ENROLLMENT_MODE_AUDIT
            }
        )
        self.enrollment_client.create_student_enrollment(
            course_id=course_id,
            username=user
        )
        self.assertDictEqual(
            request_mock.last_request.json(),
            {
                'course_details': {
                    'course_id': course_id
                },
                'mode': ENROLLMENT_MODE_AUDIT,
                'user': user
            }
        )
        self.enrollment_client.create_student_enrollment(
            course_id=course_id,
            username=user,
            enrollment_attributes=enrollment_attributes
        )
        self.assertDictEqual(
            request_mock.last_request.json(),
            {
                'enrollment_attributes': enrollment_attributes,
                'course_details': {
                    'course_id': course_id
                },
                'mode': ENROLLMENT_MODE_AUDIT,
                'user': user
            }
        )

    @patch('edx_api.enrollments.CourseEnrollments._get_enrollments_list_page')
    def test_get_enrollments(self, mock_get_enrollments_list_page):
        """
        Test get_enrollments return all enrollments.
        """
        mock_get_enrollments_list_page.side_effect = [
            ([{}, {}], 'cursor'),
            ([{}, {}], 'cursor'),
            ([{}, {}], 'cursor'),
            ([{}, {}], None)
        ]
        enrollments = list(self.enrollment_client.get_enrollments())
        assert len(enrollments) == 8

    @requests_mock.mock()
    def test_get_enrollments_list(self, mock_req):
        """
        Test get enrollments call with pagination.
        """
        mock_req.register_uri(
            'GET',
            CourseEnrollments.enrollment_list_url,
            text=json.dumps({
                'previous': None,
                'results': self.enrollments_list_json[:2],
                'next': 'http://base_url/enrl/?cursor=next-cursor'
            }),
        )
        mock_req.register_uri(
            'GET',
            '{url}?cursor=next-cursor'.format(url=CourseEnrollments.enrollment_list_url),
            text=json.dumps({
                'previous': 'http://base_url/enrl/?cursor=next-cursor',
                'results': self.enrollments_list_json[2:],
                'next': None,
            })
        )
        enrollments = list(self.enrollment_client.get_enrollments())
        assert len(enrollments) == 4

    @requests_mock.mock()
    def test_deactivate_enrollment(self, request_mock):
        """
        Test that deactivate_enrollment calls the enrollment endpoint with the correct request
        body and returns the deactivated enrollment
        """
        request_mock.post(self.enrollment_url, json=self.enrollments_json[0])
        course_id = 'course_id'
        returned_enrollment = self.enrollment_client.deactivate_enrollment(course_id=course_id)
        self.assertDictEqual(
            request_mock.last_request.json(),
            {
                'course_details': {
                    'course_id': course_id
                },
                'is_active': False
            }
        )
        assert returned_enrollment.json == self.enrollments_json[0]



"""Models Tests for the Course Detail API"""

import json
import os.path
from unittest import TestCase

from dateutil import parser

from .models import CourseDetail, CourseMode, Media


class CourseDetailTests(TestCase):
    """Tests for course detail object"""

    @classmethod
    def setUpClass(cls):
        with open(
            os.path.join(os.path.dirname(__file__), "fixtures/course_detail.json")
        ) as file_obj:
            cls.detail_json = json.loads(file_obj.read())

        cls.detail = CourseDetail(cls.detail_json)

    def test_str(self):
        """Test the __str__"""
        assert str(self.detail) == "<Course detail for course-v1:edX+DemoX+Demo_Course>"

    def test_repr(self):
        """Test the __repr__"""
        assert (
            self.detail.__repr__() == "<Course detail for course-v1:edX+DemoX+Demo_Course>"
        )

    def test_blocks_url(self):
        """Test for blocks_url property"""
        assert self.detail.blocks_url == (
            "http://192.168.33.10:8000/api/courses/v1/blocks/"
            "?course_id=course-v1%3AedX%2BDemoX%2BDemo_Course"
        )

    def test_effort(self):
        """Test for effort property"""
        assert self.detail.effort == "7 hours"

    def test_end(self):
        """Test for end property"""
        assert self.detail.end == parser.parse("2016-09-01T05:00:00Z")

    def test_enrollment_start(self):
        """Test for enrollment_start property"""
        assert self.detail.enrollment_start == parser.parse("2016-03-01T00:00:00Z")

    def test_enrollment_end(self):
        """Test for enrollment_end property"""
        assert self.detail.enrollment_end is None

    def test_course_id(self):
        """Test for course_id property"""
        assert self.detail.course_id == "course-v1:edX+DemoX+Demo_Course"

    def test_name(self):
        """Test for name property"""
        assert self.detail.name == "edX Demonstration Course"

    def test_number(self):
        """Test for number property"""
        assert self.detail.number == "DemoX"

    def test_org(self):
        """Test for org property"""
        assert self.detail.org == "edX"

    def test_short_description(self):
        """Test for short_description property"""
        assert self.detail.short_description == ""

    def test_start(self):
        """Test for start property"""
        assert self.detail.start == parser.parse("2016-03-25T05:00:00Z")

    def test_start_display(self):
        """Test for start_display property"""
        assert self.detail.start_display == "March 25, 2016"

    def test_start_type(self):
        """Test for start_type property"""
        assert self.detail.start_type == "timestamp"

    def test_overview(self):
        """Test for overview property"""
        assert self.detail.overview == (
            "<h2>About This Course</h2>\n   <p>Include your long course"
            " description here. The long course description should "
            "contain 150-400 words.</p>\n"
        )

    def test_media(self):
        """Test for media property"""
        list_media = list(self.detail.media)
        assert len(list_media) == 2
        assert (
            Media(
                type="course_image",
                url="/asset-v1:edX+DemoX+Demo_Course+type@asset+block@images_course_image.jpg",
            )
            in list_media
        )
        assert Media(type="course_video", url=None) in list_media

    def test_pacing(self):
        """Test for pacing property"""
        assert self.detail.pacing == "self"

    def test_is_self_paced(self):
        """Test for is_self_paced helper function"""
        assert self.detail.is_self_paced() is True


class CourseModeTests(TestCase):
    """Tests for course mode object"""

    @classmethod
    def setUpClass(cls):
        with open(
            os.path.join(os.path.dirname(__file__), "fixtures/course_mode.json")
        ) as file_obj:
            cls.detail_json = json.loads(file_obj.read())

        cls.detail = CourseMode(cls.detail_json[0])

    def test_str(self):
        """Test the __str__"""
        assert str(self.detail) == "<Course mode for string>"

    def test_repr(self):
        """Test the __repr__"""
        assert self.detail.__repr__() == "<Course mode for string>"

    def test_mode_slug(self):
        """Test for mode_slug property"""
        assert self.detail.mode_slug == "string"

    def test_mode_display_name(self):
        """Test for mode_display_name property"""
        assert self.detail.mode_display_name == "string"

    def test_min_price(self):
        """Test for min_price property"""
        assert self.detail.min_price == 0

    def test_currency(self):
        """Test for currency property"""
        assert self.detail.currency == "string"

    def test_expiration_datetime(self):
        """Test for expiration_datetime property"""
        assert self.detail.expiration_datetime == parser.parse("2022-08-30T17:28:48.151Z")

    def test_expiration_datetime_is_explicit(self):
        """Test for course_id property"""
        assert self.detail.expiration_datetime_is_explicit is True

    def test_description(self):
        """Test for description property"""
        assert self.detail.description == "string"



"""
edX Certificates REST API client class
"""
from requests.exceptions import HTTPError
from urllib.parse import urljoin

from edx_api.enrollments import CourseEnrollments
from .models import Certificate, Certificates


class UserCertificates(object):
    """
    edX student certificates client
    """

    def __init__(self, requester, base_url):
        """
        Args:
            requester (Requester): an authenticated objects for requests to edX
            base_url (str): string representing the base URL of an edX LMS instance
        """
        self.requester = requester
        self.base_url = base_url

    def get_student_certificate(self, username, course_id):
        """
        Returns an Certificate object with the user certificates

        Args:
            username (str): an edx user's username
            course_id (str): an edX course id.

        Returns:
            Certificate: object representing the student certificate for a course
        """
        # the request is done in behalf of the current logged in user
        resp = self.requester.get(
            urljoin(
                self.base_url,
                '/api/certificates/v0/certificates/{username}/courses/{course_key}/'.format(
                    username=username,
                    course_key=course_id
                )
            )
        )

        resp.raise_for_status()

        return Certificate(resp.json())

    def get_student_certificates(self, username, course_ids=None):
        """
        Returns an Certificates object with the user certificates

        Args:
            username (str): an edx user's username
            course_ids (list): a list of edX course ids.

        Returns:
            Certificates: object representing the student certificates for a course
        """
        # if no course ids are provided, let's get the user enrollments
        if course_ids is None:
            enrollments_client = CourseEnrollments(self.requester, self.base_url)
            enrollments = enrollments_client.get_student_enrollments()
            course_ids = list(enrollments.get_enrolled_course_ids())

        all_certificates = []
        for course_id in course_ids:
            try:
                all_certificates.append(self.get_student_certificate(username, course_id))
            except HTTPError as error:
                if error.response.status_code >= 500:
                    raise

        return Certificates(all_certificates)



"""Tests for CCX api"""
import json
import os.path
import requests

from mock import create_autospec
from .ccx import CCX


def test_create():
    """Test create method returns ccx id"""
    filename = os.path.join(os.path.dirname(__file__), 'fixtures/ccx_create_response.json')
    with open(filename) as f_obj:
        data = f_obj.read()

    mock_requester = create_autospec(requests)
    mock_requester.post.return_value.json.return_value = json.loads(data)
    ccx = CCX(mock_requester, 'https://example.org/')
    result = ccx.create('course-id', 'foo@bar.com', 100, 'test title')

    assert result == "ccx-v1:Organization+EX101+RUN-FALL2099+ccx@1"



"""
Business objects for the enrollments API
"""
from dateutil import parser

# pylint: disable=too-few-public-methods


class Enrollments(object):
    """
    The enrollments object
    """

    def __init__(self, payload):
        self.enrollments = {}
        for enrollment_json in payload:
            enrollment = Enrollment(enrollment_json)
            self.enrollments[enrollment.course_id] = enrollment

    def __str__(self):
        return "<Enrollments>"

    @property
    def enrolled_courses(self):
        """
        Returns a iterable of all the courses the user has enrolled in
        """
        return self.enrollments.values()

    def get_enrolled_courses_by_ids(self, course_id_list):
        """
        Helper function to pull the enrolled courses information for only the
        course ids specified in course_id_list

        Args:
            course_id_list (List): list of course IDs

        Returns:
            generator of Enrollment objects
        """
        if not isinstance(course_id_list, (list, tuple)):
            raise ValueError('course_id_list should be an instance of list or tuple')
        for course_id in course_id_list:
            yield self.enrollments.get(course_id)

    def get_enrolled_course_ids(self, filter_by_id_list=None):
        """
        Helper function that extracts the course id only from the course IDs
        Args:
            filter_by_id_list (List): list of course IDs

        Returns:
            iterable of strings representing course IDs
        """
        all_ids = self.enrollments.keys()
        if filter_by_id_list is not None:
            all_ids = set(all_ids).intersection(set(filter_by_id_list))
        return all_ids

    def is_enrolled_in(self, course_id):
        """
        Helper to check if the provided course id is in the enrollments

        Args:
            course_id (str): a string representing a course id

        Returns:
            bool: whether an enrollment for the provided course id exists
        """
        return course_id in self.enrollments

    def get_enrollment_for_course(self, course_id):
        """
        Helper to get a specific enrollment for a course_id
        """
        return self.enrollments.get(course_id)


class Enrollment(object):
    """
    Single enrollment object representation
    """
    def __init__(self, json):
        self.json = json

    def __str__(self):
        return "<Enrollment for user {user} in course {course}>".format(
            user=self.user,
            course=self.course_id
        )

    @property
    def course_id(self):
        """Shortcut for a nested property"""
        return self.course_details.course_id or self.json.get('course_id')

    @property
    def created(self):
        """Returns a datetime object of the enrollment timestamp"""
        try:
            return parser.parse(self.json.get('created'))
        except (AttributeError, TypeError):
            return None

    @property
    def mode(self):
        """Returns the enrollment mode of the user in this course"""
        return self.json.get('mode')

    @property
    def is_active(self):
        """Returns whether the enrollment is currently active."""
        return self.json.get('is_active') is True

    @property
    def course_details(self):
        """Returns a CourseDetails object"""
        return CourseDetails(self.json.get('course_details', {}))

    @property
    def user(self):
        """Returns the username of the user."""
        return self.json.get('user')

    @property
    def is_verified(self):
        """
        Checks if the mode is "verified"
        """
        return self.mode == 'verified'


class CourseDetails(object):
    """
    Course enrollment info
    """
    def __init__(self, json):
        self.json = json

    def __str__(self):
        return "<Enrollment details for course {}>".format(self.course_id)

    @property
    def course_id(self):
        """Returns the unique identifier for the course."""
        return self.json.get('course_id')

    @property
    def course_start(self):
        """
        Returns the date and time when the course opens.
        If None, the course opens immediately when it is created.
        """
        try:
            return parser.parse(self.json.get('course_start'))
        except (AttributeError, TypeError):
            return None

    @property
    def course_end(self):
        """
        Returns the date and time when the course closes.
        If None, the course never ends.
        """
        try:
            return parser.parse(self.json.get('course_end'))
        except (AttributeError, TypeError):
            return None

    @property
    def enrollment_start(self):
        """
        Returns the date and time when users can begin enrolling in the course.
        If None, enrollment opens immediately when the course is created.
        """
        try:
            return parser.parse(self.json.get('enrollment_start'))
        except (AttributeError, TypeError):
            return None

    @property
    def enrollment_end(self):
        """
        Returns the date and time after which users cannot enroll for the course.
        If None, the enrollment period never ends.
        """
        try:
            return parser.parse(self.json.get('enrollment_end'))
        except (AttributeError, TypeError):
            return None

    @property
    def invite_only(self):
        """
        Returns a boolean indicating whether students
        must be invited to enroll in the course.
        """
        return self.json.get('invite_only') is True

    @property
    def course_modes(self):
        """
        Returns a generator of data about the enrollment
        modes supported for the course.
        """
        for course_mode_json in self.json.get('course_modes', []):
            yield CourseMode(course_mode_json)


class CourseMode(object):
    """
    Course enrollment mode
    """
    def __init__(self, json):
        self.json = json

    def __str__(self):
        return "<Enrollment mode {}>".format(self.slug)

    @property
    def currency(self):
        """Returns the currency of the listed prices."""
        return self.json.get('currency')

    @property
    def description(self):
        """Returns a description of this mode."""
        return self.json.get('description')

    @property
    def expiration_datetime(self):
        """
        Returns the date and time after which users
        cannot enroll in the course in this mode.
        """
        try:
            return parser.parse(self.json.get('expiration_datetime'))
        except (AttributeError, TypeError):
            return None

    @property
    def min_price(self):
        """Returns the minimum price for which a user can enroll in this mode."""
        return self.json.get('min_price')

    @property
    def name(self):
        """Returns the full name of the enrollment mode."""
        return self.json.get('name')

    @property
    def slug(self):
        """Returns the short name for the enrollment mode."""
        return self.json.get('slug')

    @property
    def suggested_prices(self):
        """Returns a list of suggested prices for this enrollment mode."""
        return self.json.get('suggested_prices')



"""
Business objects for the Grades API
"""
from collections.abc import Iterable
# pylint: disable=too-few-public-methods


class CurrentGrades(object):
    """
    Current Grades object representation
    """
    def __init__(self, current_grade_list):
        if not isinstance(current_grade_list, Iterable):
            raise TypeError('CurrentGrades needs an Iterable object')

    @property
    def all_current_grades(self):
        """Helper property to return all the CurrentGrade objects"""
        return self.current_grades.values()  # pylint: disable=no-member


class CurrentGradesByCourse(CurrentGrades):
    """
    Represents the current grades for a specific course
    """
    def __init__(self, current_grade_list):
        """
        Args:
            current_grade_list (list): A list of the CurrentGrade objects
        """
        super(CurrentGradesByCourse, self).__init__(current_grade_list)
        self.course_id = None
        self.current_grades = {}
        for current_grade in current_grade_list:
            if not isinstance(current_grade, CurrentGrade):
                raise ValueError("Only CurrentGrade objects are allowed")
            if self.course_id is None:
                self.course_id = current_grade.course_id
            if self.course_id is not None and current_grade.course_id != self.course_id:
                raise ValueError("Only CurrentGrade objects for the same course are allowed")
            self.current_grades[current_grade.username] = current_grade

    def __str__(self):
        return "<Current Grades for course {course_id}>".format(
            course_id=self.course_id
        )

    @property
    def all_usernames(self):
        """Helper property to return all the usernames of the current grades"""
        return self.current_grades.keys()


class CurrentGradesByUser(CurrentGrades):
    """
    Represents the current grades for a specific user
    """
    def __init__(self, current_grade_list):
        """
        Args:
            current_grade_list (list): A list of the CurrentGrade objects
        """
        super(CurrentGradesByUser, self).__init__(current_grade_list)
        self.username = None
        self.current_grades = {}
        for current_grade in current_grade_list:
            if not isinstance(current_grade, CurrentGrade):
                raise ValueError("Only CurrentGrade objects are allowed")
            self.current_grades[current_grade.course_id] = current_grade

    def __str__(self):
        return "<Current Grades for user {username}>".format(
            username=self.username
        )

    @property
    def all_course_ids(self):
        """Helper property to return all the course ids of the current grades"""
        return self.current_grades.keys()

    def get_current_grade(self, course_id):
        """Returns the current grade for the given course id"""
        return self.current_grades.get(course_id)


class CurrentGrade(object):
    """
    Single current grade object representation
    """
    def __init__(self, json):
        self.json = json

    def __str__(self):
        return "<Current Grade for user {user} in course {course}>".format(
            user=self.username,
            course=self.course_id
        )

    @property
    def course_id(self):
        """Shortcut for a nested property"""
        return self.json.get('course_id')

    @property
    def email(self):
        """Returns email of the user"""
        return self.json.get('email')

    @property
    def username(self):
        """Returns the username of the user."""
        return self.json.get('username')

    @property
    def passed(self):
        """Whether the user has passed the course"""
        return self.json.get('passed')

    @property
    def percent(self):
        """
        Returns a decimal representation between
        0 and 1 of the student grade for the course
        """
        return self.json.get('percent')

    @property
    def letter_grade(self):
        """
        Returns a letter grade as defined in the
        edX grading_policy (e.g. 'A' 'B' 'C' for 6.002x) or None
        """
        return self.json.get('letter_grade')



"""
Test handling of responses from grades api.
"""

import json
import os
from unittest import TestCase
from urllib.parse import urljoin

import requests_mock

from edx_api import enrollments, grades
from edx_api.client import EdxApi


class GradesApiTestCase(TestCase):
    """
    Test handling of mocked API responses in Grades API client.
    """

    base_url = "https://edx.example.com"

    def setUp(self):
        super(GradesApiTestCase, self).setUp()

        with open(
            os.path.join(
                os.path.dirname(__file__),
                "../enrollments/fixtures/user_enrollments.json",
            )
        ) as file:  # pylint: disable=redefined-builtin
            self.enrollment_data = json.load(file)
        self.enrollment_url = urljoin(
            "https://edx.example.com", enrollments.CourseEnrollments.enrollment_url
        )
        self.client = EdxApi({"access_token": "opensesame"}, "https://edx.example.com")

    @requests_mock.mock()
    def test_api_return_value(self, mock_req):
        """
        Verify that the object returned from the current_grades endpoints is a CurrentGrades.
        """
        mock_req.get(
            requests_mock.ANY,
            text=json.dumps(self.get_grades_data("course_grades_hawthorn.json")),
        )
        mock_req.get(self.enrollment_url, text=json.dumps(self.enrollment_data))
        response = self.client.current_grades.get_course_current_grades(
            "course-v1:edX+DemoX+Demo_Course"
        )
        assert isinstance(response, grades.CurrentGradesByCourse)

    @requests_mock.mock()
    def test_ironwood_api(self, mock_req):
        """
        Verify that the api can handle the ironwood version of the api.
        """
        mock_req.get(
            requests_mock.ANY,
            text=json.dumps(self.get_grades_data("course_grades_ironwood_p2.json")),
        )
        mock_req.get(self.enrollment_url, text=json.dumps(self.enrollment_data))
        grades_response = self.client.current_grades.get_course_current_grades(
            "course-v1:edX+DemoX+Demo_Course"
        )
        assert isinstance(grades_response, grades.CurrentGradesByCourse)
        self.assertEqual(len(grades_response.current_grades), 2)

    @requests_mock.mock()
    def test_paginated_ironwood_api(self, mock_req):
        """
        Verify that the ironwood api handles paginated data properly
        """
        mock_req.get(
            requests_mock.ANY,
            [
                {
                    "text": json.dumps(
                        self.get_grades_data("course_grades_ironwood_p1.json")
                    )
                },
                {
                    "text": json.dumps(
                        self.get_grades_data("course_grades_ironwood_p2.json")
                    )
                },
            ],
        )
        mock_req.get(self.enrollment_url, text=json.dumps(self.enrollment_data))
        grades_response = self.client.current_grades.get_course_current_grades(
            "course-v1:edX+DemoX+Demo_Course"
        )
        self.assertIsInstance(grades_response, grades.CurrentGradesByCourse)
        self.assertEqual(len(grades_response.current_grades), 4)

    @staticmethod
    def get_grades_data(filename):
        """
        Return the JSON data from the named fixture.
        """
        with open(
            os.path.join(os.path.dirname(__file__), "fixtures", filename)
        ) as file:  # pylint: disable=redefined-builtin
            return json.load(file)



"""Course Structure API"""
from urllib.parse import urljoin

from .models import Structure


class CourseStructure(object):
    """
    API Client to interface with the course structure API.
    """
    def __init__(self, requester, base_url):
        self.requester = requester
        self.base_url = base_url

    def course_blocks(self, course_id, username):
        """
        Fetches course blocks.

        Args:
            course_id (str): An edx course id.
            username (str): username of the user to query for (can reveal hidden
                            modules)

        Returns:
            Structure
        """
        resp = self.requester.get(
            urljoin(self.base_url, '/api/courses/v1/blocks/'),
            params={
                "depth": "all",
                "username": username,
                "course_id": course_id,
                "requested_fields": "children,display_name,id,type,visible_to_staff_only",
            })

        resp.raise_for_status()

        return Structure(resp.json())



"""Client for user_info API"""
from urllib.parse import urljoin

from .models import Info


class UserInfo(object):
    """
    edX user info client
    """

    def __init__(self, requester, base_url):
        """
        Args:
            requester (Requester): an authenticated objects for requests to edX
            base_url (str): string representing the base URL of an edX LMS instance
        """
        self.requester = requester
        self.base_url = base_url

    def get_user_info(self):
        """
        Returns a UserInfo object for the logged in user.

        Returns:
            UserInfo: object representing the student current grades
        """
        # the request is done in behalf of the current logged in user
        resp = self.requester.get(
            urljoin(
                self.base_url,
                '/api/mobile/v0.5/my_user_info'
            )
        )

        resp.raise_for_status()

        return Info(resp.json())

    def update_user_name(self, username, full_name):
        """
        Updates full name of the user

        Args:
            username (str): Username of the Application user
            full_name (str): Full name that will replace the user's existing full name

        Returns:
            UserInfo: Object representing the edX user
        """
        request_data = dict(name=full_name)

        # the request is done on behalf of the current logged in user
        self.requester.headers.update(
            {
                "Accept": "application/json, text/javascript, */*; q=0.01",
                "X-Requested-With": "XMLHttpRequest",
                "Content-Type": "application/merge-patch+json",
            }
        )
        resp = self.requester.patch(
            urljoin(
                self.base_url,
                '/api/user/v1/accounts/{username}'.format(username=username)
            ),
            json=request_data)
        resp.raise_for_status()

        return Info(resp.json())



# -*- encoding: utf-8 -*-
"""
Python setup file for the eds_api app.
"""
import os
from setuptools import setup, find_packages

import edx_api

# pylint: disable=invalid-name
dev_requires = [
    'flake8',
]

install_requires = open('requirements.txt').read().splitlines()


def read(filename):
    """Helper function to read bytes from file"""
    try:
        return open(os.path.join(os.path.dirname(__file__), filename)).read()
    except IOError:
        return ''


setup(
    name="edx-api-client",
    version=edx_api.__version__,
    description=read('DESCRIPTION'),
    long_description=read('README.rst'),
    license='The AGPL License',
    platforms=['OS Independent'],
    keywords='edx, rest api',
    author='MIT Office of Digital Learning',
    author_email='mitx-devops@mit.edu',
    url="https://github.com/mitodl/edx-api-client",
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'dev': dev_requires,
    },
)



"""Models for user_info client"""


class Info(object):
    """
    Information about user
    """
    def __init__(self, json):
        self.json = json

    def __str__(self):
        return "<User info for user {user}>".format(
            user=self.username,
        )

    @property
    def username(self):
        """Returns the username of the user."""
        return self.json.get('username')

    @property
    def email(self):
        """Email for the user"""
        return self.json.get('email')

    @property
    def name(self):
        """Name for the user"""
        return self.json.get('name')

    @property
    def user_id(self):
        """Id for the user"""
        return self.json.get('id')



"""Business objects for the course structure API"""


class Structure(object):
    """
    The course structure object, which represents a tree of nodes.
    """
    def __init__(self, payload):
        self.payload = payload

    @property
    def blocks(self):
        """Returns all the blocks in the structure"""
        for block_id in self.payload['blocks'].keys():
            yield Block(block_id, self.payload)

    @property
    def _root_id(self):
        return self.payload['root']

    @property
    def root(self):
        """Returns the root node in the course"""
        root_id = self._root_id
        return Block(root_id, self.payload)

    def __str__(self):
        return 'Structure for {}'.format(self._root_id)


class Block(object):
    """
    Represents a single block within the course structure.
    """
    def __init__(self, block_id, json):
        self.json = json
        self.block_id = block_id

    def _get_block(self, block_id=None):
        if not block_id:
            block_id = self.block_id
        return self.json['blocks'][block_id]

    @property
    def children(self):
        """Returns blocks for each of this blocks children"""
        children = []
        for child in self._get_block()['children']:
            children.append(Block(child, self.json))
        return children

    @property
    def visible(self):
        """Returns whether the block is visible to non-staff"""
        # False, because if we're a non-staff user, we don't get the
        # visibile_to_staff_only field, it seems.
        return not self._get_block().get('visible_to_staff_only', False)

    @property
    def title(self):
        """Returns title"""
        return self._get_block()['display_name']

    def __str__(self):
        return self.block_id



"""edX api client"""
# pylint: disable=fixme
import requests

from . import DEFAULT_TIME_OUT
from .ccx import CCX
from .certificates import UserCertificates
from .course_detail import CourseDetails, CourseModes
from .course_structure import CourseStructure
from .enrollments import CourseEnrollments
from .email_settings import EmailSettings
from .grades import UserCurrentGrades
from .user_info import UserInfo


class EdxApi(object):
    """
    A client for speaking with edX.
    """
    def __init__(self, credentials, base_url='https://courses.edx.org/', timeout=DEFAULT_TIME_OUT):
        if 'access_token' not in credentials:
            raise AttributeError(
                "Due to a lack of support for Client Credentials Grant in edX,"
                " you must specify the access token."
            )

        self.base_url = base_url
        self.credentials = credentials
        self.timeout = timeout

    def get_requester(self):
        """
        Returns an object to make authenticated requests. See python `requests` for the API.
        """
        # TODO(abrahms): Perhaps pull this out into a factory function for
        # generating an EdxApi instance with the proper requester & credentials.
        session = requests.session()
        session.headers.update({
            'Authorization': 'Bearer {}'.format(self.credentials['access_token']),
            'X-EdX-Api-Key': self.credentials.get('api_key'),
        })

        old_request = session.request

        def patched_request(*args, **kwargs):
            """
            adds timeout param to session.request
            """
            return old_request(*args, timeout=self.timeout, **kwargs)

        session.request = patched_request
        return session

    @property
    def course_structure(self):
        """Course Structure API"""
        return CourseStructure(self.get_requester(), self.base_url)

    @property
    def course_detail(self):
        """Course Detail API"""
        return CourseDetails(self.get_requester(), self.base_url)

    @property
    def course_mode(self):
        """Course Detail API"""
        return CourseModes(self.get_requester(), self.base_url)

    @property
    def enrollments(self):
        """Course Enrollments API"""
        return CourseEnrollments(self.get_requester(), self.base_url)

    @property
    def ccx(self):
        """CCX API"""
        return CCX(self.get_requester(), self.base_url)

    @property
    def email_settings(self):
        """Email Settings API"""
        return EmailSettings(self.get_requester(), self.base_url)

    @property
    def certificates(self):
        """Certificates API"""
        return UserCertificates(self.get_requester(), self.base_url)

    @property
    def current_grades(self):
        """Current Grades API"""
        return UserCurrentGrades(self.get_requester(), self.base_url)

    @property
    def user_info(self):
        """User info API"""
        return UserInfo(self.get_requester(), self.base_url)



"""
API client for interacting with change email settings
"""
import logging
from urllib import parse

log = logging.getLogger(__name__)


class EmailSettings(object):
    """
    API client for interacting with email settings
    """
    api_url = "/api/change_email_settings"

    def __init__(self, requester, base_url):
        self.requester = requester
        self.base_url = base_url

    def change_settings(self, payload):
        """
        Execute the client request to edX endpoint
        Args:
            payload (dict): request payload
        Returns:
            JSON response (dict)
        """
        response = self.requester.post(
            parse.urljoin(self.base_url, self.api_url),
            json=payload
        )

        try:
            response.raise_for_status()
        except Exception:
            log.error(response.json())
        return response.json().get("success", False)

    def subscribe(self, course_id):
        """
        Subscribe the user to receive all course emails
        Args:
            course_id (int): Corresponding edx course id
        """
        payload = {
            "course_id": course_id,
            "receive_emails": "on",
        }
        return self.change_settings(payload)

    def unsubscribe(self, course_id):
        """
        Unsubscribe the user from receiving course emails
        Args:
            course_id (int): Corresponding edx course
        """
        payload = {
            "course_id": course_id,
        }
        return self.change_settings(payload)



"""
Business objects for the course detail API
"""
from collections import namedtuple

from dateutil import parser

Media = namedtuple("Media", ["type", "url"])


class CourseDetail(object):
    """
    The course detail object
    """

    def __init__(self, payload):
        self.json = payload

    def __str__(self):
        return "<Course detail for {}>".format(self.course_id)

    def __repr__(self):
        return self.__str__()

    @property
    def blocks_url(self):
        """Used to fetch the course blocks"""
        return self.json.get("blocks_url")

    @property
    def effort(self):
        """A textual description of the weekly hours of effort expected in the course."""
        return self.json.get("effort")

    @property
    def end(self):
        """Date the course ends"""
        try:
            return parser.parse(self.json.get("end"))
        except (AttributeError, TypeError):
            return None

    @property
    def enrollment_start(self):
        """Date enrollment begins"""
        try:
            return parser.parse(self.json.get("enrollment_start"))
        except (AttributeError, TypeError):
            return None

    @property
    def enrollment_end(self):
        """Date enrollment ends"""
        try:
            return parser.parse(self.json.get("enrollment_end"))
        except (AttributeError, TypeError):
            return None

    @property
    def course_id(self):
        """
        A unique identifier of the course; a serialized representation
        of the opaque key identifying the course.
        """
        return self.json.get("id")

    @property
    def name(self):
        """Name of the course"""
        return self.json.get("name")

    @property
    def number(self):
        """Catalog number of the course"""
        return self.json.get("number")

    @property
    def org(self):
        """Name of the organization that owns the course"""
        return self.json.get("org")

    @property
    def short_description(self):
        """A textual description of the course"""
        return self.json.get("short_description")

    @property
    def start(self):
        """Date the course begins"""
        try:
            return parser.parse(self.json.get("start"))
        except (AttributeError, TypeError):
            return None

    @property
    def start_display(self):
        """Readably formatted start of the course"""
        return self.json.get("start_display")

    @property
    def start_type(self):
        """
        Hint describing how `start_display` is set. One of:
            * `"string"`: manually set
            * `"timestamp"`: generated form `start` timestamp
            * `"empty"`: the start date should not be shown
        """
        return self.json.get("start_type")

    @property
    def overview(self):
        """
        A possibly verbose HTML textual description of the course.
        Note: this field is only included in the Course Detail view, not
        the Course List view.
        """
        return self.json.get("overview")

    @property
    def media(self):
        """Contains named media items"""
        for media_type, url_dict in self.json.get("media", {}).items():
            yield Media(type=media_type, url=url_dict.get("uri"))

    @property
    def pacing(self):
        """
        Pacing of a course. Possible values are ("self" or "instructor")
        """
        return self.json.get("pacing")

    def is_self_paced(self):
        """
        Helper function to check if a course is self paced
        Note: This property is not part of course detail API, It's calculated on base of course pacing
        """
        return self.pacing == "self"


class CourseMode(object):
    """
    The course mode object
    """

    def __init__(self, payload):
        self.json = payload

    def __str__(self):
        return "<Course mode for {}>".format(self.course_id)

    def __repr__(self):
        return self.__str__()

    @property
    def course_id(self):
        """The course ID associated with the course mode."""
        return self.json.get("course_id")

    @property
    def mode_slug(self):
        """The short name for the course mode."""
        return self.json.get("mode_slug")

    @property
    def mode_display_name(self):
        """The verbose name for the course mode."""
        return self.json.get("mode_display_name")

    @property
    def min_price(self):
        """The minimum price for which a user can enroll in this mode."""
        return self.json.get("min_price")

    @property
    def currency(self):
        """The currency of the listed prices."""
        return self.json.get("currency")

    @property
    def expiration_datetime(self):
        """The date and time after which users cannot enroll in the course in this mode"""
        try:
            return parser.parse(self.json.get("expiration_datetime"))
        except (AttributeError, TypeError, IndexError, KeyError):
            return None

    @property
    def expiration_datetime_is_explicit(self):
        """
        Whether the expiration_datetime field was
        explicitly set
        """
        return self.json.get("expiration_datetime_is_explicit")

    @property
    def description(self):
        """A description of this mode"""
        return self.json.get("description")
